/*
 * i2c-emma3sx.c
 *
 * I2C adapter for the EMMA3Sx I2C bus access.
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <config.h>

#include <common.h>
#include <asm/io.h>
#include <asm/arch-emma/emma.h>
#include <i2c.h>

#define I2C_M_RD	0x01
#define I2C_M_WR	0x00

#define I2C_OK		0
#define I2C_NOK		1
#define I2C_NACK	2
#define I2C_NOK_LA	3		/* Lost arbitration */
#define I2C_NOK_TOUT	4	/* time out */

#define I2C_TIMEOUT 1		/* 1 second */

/* I2C registers */
/* control */
#define I2C_CON_START_COND	(0x9a)
#define I2C_CON_SEND_ACK	(0xbc)
#define I2C_CON_SEND_NACK	(0xb8)
#define I2C_CON_STOP_COND	(0x99)

/* status */
#define I2C_STS_START_DET	(0x01)
#define I2C_STS_STOP_DET	(0x02)
#define I2C_STS_ACK_DET		(0x04)
#define I2C_STS_TRANSMIT	(0x08)
#define I2C_STS_ABT_LOSS	(0x40)
#define I2C_STS_MASTER		(0x80)

/* clock select */
#define I2C_CLK_MASK	(0x8b)

/* interrupt */
#define I2C_INT_TRANS	(0x01)

struct emma_i2c {
	u32		piic_iicc;	/* Control */
	u32		pad0[3];
	u32		piic_iics;	/* Status */
	u32		pad1[3];
	u32		piic_iiccl;	/* Clock Select */
	u32		pad2[3];
	u32		piic_sva;	/* Slave Address */
	u32		pad3[3];
	u32		piic_iic;	/* Data */
	u32		pad4[3];
	u32		piic_int;	/* Interrupt Status */
	u32		pad5[3];
	u32		piic_intm0;	/* Interrupt Mask0 */
	u32		piic_intm1;	/* Interrupt Mask1 */
	u32		pad6[6];
	u32		piic_pcnt;	/* Pin Control */
};

static struct emma_i2c *i2c;
static unsigned int i2c_reg_base[CONFIG_SYS_MAX_I2C_BUS];
static unsigned int i2c_clock;
static unsigned int i2c_bus_num; /* I2C Bus */

static void i2c_set_clock(u32 val)
{
	u32 reg;

	if (val >= 375000)
		reg = 0x0b;		/* 375KHz */
	else if ((val >= 321000) && (val < 375000))
		reg = 0x8a;		/* 321KHz */
	else if ((val >= 281000) && (val < 321000))
		reg = 0x0a;		/* 281KHz */
	else if ((val >= 200000) && (val < 281000))
		reg = 0x09;		/* 200KHz */
	else if ((val >= 140000) && (val < 200000))
		reg = 0x08;		/* 140KHz */
	else if ((val >= 117000) && (val < 140000))
		reg = 0x83;		/* 17KHz */
	else if ((val >= 10000) && (val < 117000))
		reg = 0x03;		/* 102KHz */
	else if ((val >=  76000) && (val < 100000))
		reg = 0x00;		/*  76KHz */
	else
		reg = 0x01;		/*  40KHz */

	writel(reg, &i2c->piic_iiccl);
}

static int wait_for_xfer(void)
{
	int i, result;

	i = I2C_TIMEOUT * 100000;
	while (!(readl(&i2c->piic_int) & I2C_INT_TRANS) && (i > 0)) {
		udelay(10);
		i--;
	}

	if (i == 0) {
		result = I2C_NOK_TOUT;
	} else {
		writel(0x00, &i2c->piic_int);
		result = I2C_OK;
	}

	return result;
}

static void i2c_bus_reset(void)
{
	int i;

	writel(I2C_CON_STOP_COND, &i2c->piic_iicc);

	for (i = 0; i < 10; i++) {
		/* SCL=High */
		writel(0x03, &i2c->piic_pcnt);
		udelay(100);
		/* SCL=Low */
		writel(0x01, &i2c->piic_pcnt);
		udelay(100);
		if (readl(&i2c->piic_iiccl) & 0x10) {	/* SDA=High? */
			break;
		}
	}
	/* generate manual stop conditions */
	/* SDA=Low */
	writel(0x04, &i2c->piic_pcnt);
	udelay(100);
	/* SCL=High */
	writel(0x03, &i2c->piic_pcnt);
	udelay(100);
	/* SDA=High */
	writel(0x0c, &i2c->piic_pcnt);
	udelay(100);

	/* disabled manual control mode */
	writel(0x00, &i2c->piic_pcnt);
	udelay(100);

	writel(0x00, &i2c->piic_int);
}

/*
 * cmd_type is 0 for write, 1 for read.
 *
 * addr_len can take any value from 0-255, it is only limited
 * by the char, we could make it larger if needed. If it is
 * 0 we skip the address write cycle.
 */
static int i2c_transfer(unsigned char cmd_type,
		 unsigned char chip,
		 unsigned char addr[],
		 unsigned char addr_len,
		 unsigned char data[], unsigned short data_len)
{
	int result;
	int cnt;
	u32 status;

	if (data == 0 || data_len == 0) {
		/*Don't support data transfer of no length or to address 0 */
		printf("i2c_transfer: bad call\n");
		return I2C_NOK;
	}

	/* send start conditions */
	writel(I2C_CON_START_COND, &i2c->piic_iicc);

	/* send i2c target address */
	if ((cmd_type == I2C_M_RD) && (addr_len == 0))
		writel((chip | 0x01), &i2c->piic_iic);
	else
		writel((chip & ~0x01), &i2c->piic_iic);

	result = wait_for_xfer();
	if (result == I2C_OK) {
		status = readl(&i2c->piic_iics);
		if ((status & I2C_STS_ACK_DET) == 0)
			result = I2C_NACK;
	}

	/* sende slave device address */
	cnt = 0;
	while ((cnt < addr_len) && (result == I2C_OK)) {
		writel(addr[cnt], &i2c->piic_iic);
		result = wait_for_xfer();
		cnt++;
	}

	if ((result == I2C_OK) && (cmd_type == I2C_M_RD) && (addr_len > 0)) {
		/* perform restart condition */
		writel(I2C_CON_START_COND, &i2c->piic_iicc);
		/* send i2c target address */
		writel((chip | 0x01), &i2c->piic_iic);
		result = wait_for_xfer();
		if (result == I2C_OK) {
			status = readl(&i2c->piic_iics);
			if ((status & I2C_STS_ACK_DET) == 0)
				result = I2C_NACK;
		}
	}

	/* data transaction */
	cnt = 0;
	while ((cnt < data_len) && (result == I2C_OK)) {
		status = readl(&i2c->piic_iics);
		if ((status & I2C_STS_TRANSMIT) == I2C_STS_TRANSMIT) {
			/* transmit */
			writel(data[cnt], &i2c->piic_iic);
			result = wait_for_xfer();
		} else {	/* receive */
			if ((cnt + 1) == data_len)
				writel(I2C_CON_SEND_NACK, &i2c->piic_iicc);
			else
				writel(I2C_CON_SEND_ACK, &i2c->piic_iicc);
			result = wait_for_xfer();
			data[cnt] = readl(&i2c->piic_iic);
		}
		cnt++;
	}

	/* send stop condition */
	writel(I2C_CON_STOP_COND, &i2c->piic_iicc);
	wait_for_xfer();

	return result;
}

void i2c_init(int speed, int slaveaddr)
{
	int i;
	int result;

#ifdef CONFIG_SYS_I2C_INIT_BOARD
	/*
	 * Call board specific i2c bus reset routine before accessing the
	 * environment, which might be in a chip on that bus. For details
	 * about this problem see doc/I2C_Edge_Conditions.
	 */
	i2c_init_board();
#endif

	for (i = 0; i < CONFIG_SYS_MAX_I2C_BUS; i++) {
		switch (i) {
		case 0:
			i2c_reg_base[i] = KSEG1ADDR(PIIC0_BASE);
			break;
#if (CONFIG_SYS_MAX_I2C_BUS > 1)
		case 1:
			i2c_reg_base[i] = KSEG1ADDR(PIIC1_BASE);
			break;
#endif
#if (CONFIG_SYS_MAX_I2C_BUS > 2)
		case 2:
			i2c_reg_base[i] = KSEG1ADDR(PIIC2_BASE);
			break;
#endif
		}
		i2c = (struct emma_i2c *)i2c_reg_base[i];

		writel(0x00, &i2c->piic_iicc);
		writel(0x00, &i2c->piic_intm0);
		writel(0x00, &i2c->piic_intm1);
		writel(0x00, &i2c->piic_int);

		i2c_clock = CONFIG_SYS_I2C_SPEED;
		i2c_set_clock(i2c_clock);

		if ((readl(&i2c->piic_iiccl) & 0x30) == 0x20) {	/* SCL=High, SDA=Low */
			printf("i2c bus reset peformed!\n");
			i2c_bus_reset();
		}

		/* send stop condition */
		writel(I2C_CON_STOP_COND, &i2c->piic_iicc);
		result = wait_for_xfer();
	}

	i2c = (struct emma_i2c *)i2c_reg_base[i2c_bus_num];
}

/*
 * Probe the given I2C chip address.  Returns 0 if a chip responded,
 * not 0 on failure.
 */
int i2c_probe(uchar chip)
{
	uchar buf[1];

	buf[0] = 0;

	/*
	 * What is needed is to send the chip address and verify that the
	 * address was <ACK>ed (i.e. there was a chip at that address which
	 * drove the data line low).
	 */
	return (i2c_transfer(1, (chip << 1), 0, 0, buf, I2C_M_RD) != 0);
}

int i2c_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	uchar xaddr[4];
	int ret;

	if (alen > 4) {
		printf("I2C read: addr len %d not supported\n", alen);
		return 1;
	}

	if (alen > 0) {
		xaddr[0] = (addr >> 24) & 0xFF;
		xaddr[1] = (addr >> 16) & 0xFF;
		xaddr[2] = (addr >> 8) & 0xFF;
		xaddr[3] = addr & 0xFF;
	}

	ret = i2c_transfer(I2C_M_RD, (chip << 1), &xaddr[4 - alen], alen,
			buffer, len);
	if (ret != I2C_OK) {
		printf("I2C read: failed %d\n", ret);
		return ret;
	}
	return 0;
}

int i2c_write(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	uchar xaddr[4];

	if (alen > 4) {
		printf("I2C write: addr len %d not supported\n", alen);
		return 1;
	}

	if (alen > 0) {
		xaddr[0] = (addr >> 24) & 0xFF;
		xaddr[1] = (addr >> 16) & 0xFF;
		xaddr[2] = (addr >> 8) & 0xFF;
		xaddr[3] = addr & 0xFF;
	}
	return (i2c_transfer
		(I2C_M_WR, (chip << 1), &xaddr[4 - alen], alen, buffer,
		 len) != 0);
}

int i2c_set_bus_num(unsigned int bus)
{
	if ((bus < 0) || (bus >= CONFIG_SYS_MAX_I2C_BUS))
		return -1;

	i2c_bus_num = bus;
	i2c = (struct emma_i2c *)i2c_reg_base[i2c_bus_num];

	return 0;
}

unsigned int i2c_get_bus_num(void)
{
	return i2c_bus_num;
}

int i2c_set_bus_speed(unsigned int speed)
{
	if (speed < 0)
		return -1;

	i2c_set_clock(speed);
	i2c_clock = speed;

	return 0;
}

unsigned int i2c_get_bus_speed(void)
{
	return i2c_clock;
}

