/*
 *  drivers/mtd/nand/nand_emma.c
 *
 *  Author: NEC Electronics Corporation
 *
 *  Copyright (C) 2006-2008 NEC Electronics Corporation
 *  Copyright (C) 2004 NEC Micro Systems, Ltd.
 *
 *  Derived from drivers/mtd/nand/edb7312.c
 *       Copyright (C) 2002 Marius Groeger (mag@sysgo.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device
 *   found on the EMMA board which utilizes the NEC EMMA3PF part.
 */

#include <common.h>
#include <linux/mtd/nand.h>
#include <malloc.h>
#include <asm/io.h>
#include <asm/errno.h>

/*
 * Values specific to the EMMA board
 */
#define NAND_ROM_STATUS_BUSY		(0x00000001)

#define NAND_CS_BIT			(0x1 << CONFIG_SYS_NAND_CS)

#define NAND_CS_SHIFT			(8)
#define NAND_CMD0_SHIFT			(16)
#define NAND_CMD1_SHIFT			(24)

#define NAND_SEQ_MAX_SIZE		(32*1024)
#define NAND_SEQ_DMA_SIZE		(16*1024)

#define NAND_SEQ_READ			0
#define NAND_SEQ_WRITE			1

#define EMMA_NAND_NAME			"nand_emma"

/*
 * MTD structure for EMMA board
 */

static struct emma_nand_info_t{
	void			(*seq_cmdfunc)(struct mtd_info *mtd,
					       unsigned command,
					       int page_addr,
					       int size_bit);
} emma_nand_info;

/*
 * Debug
 */
#define NAND_DEBUG			0

#if NAND_DEBUG

static int nand_debug;

#define DPRINTK(level, args...)				\
		do {					\
			if ((level) <= nand_debug)	\
				printk(KERN_INFO args);	\
		} while (0)
#else
#define DPRINTK(level, args...)
#endif				/* NAND_DEBUG */

/*
 * Module stuff
 */

#if 1 /* Add for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/14 refs #34 @ vanhalen */
static uint8_t scan_ff_pattern[] = { 0xff, 0xff };

static struct nand_bbt_descr emma_smallpage_flashbased = {
	.options = NAND_BBT_SCAN2NDPAGE,
	.offs = 8, /* was 5 (Spareデータ内の spare[5] から spare[8] へ変更) */
	.len = 1,
	.pattern = scan_ff_pattern
};

static struct nand_bbt_descr emma_largepage_flashbased = {
	.options = NAND_BBT_SCAN2NDPAGE,
	.offs = 32, /* was 0 (Spareデータ内の spare[0],spare[1] から spare[32],spare[33] へ変更) */
	.len = 2,
	.pattern = scan_ff_pattern
};
#endif

/* #if CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
static struct nand_ecclayout emma_nand_oob_64 = {
	.eccbytes = 24,
	.eccpos = {
		    1,  0,  2,  5,  4,  6,  9,  8,
		   10, 13, 12, 14, 17, 16, 18, 21,
		   20, 22, 25, 24, 26, 29, 28, 30},
#if 1 /* Mod for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/14 refs #34 @ vanhalen */
	/* BadBlockマーカー位置を spare[32],spare[33] に変えたので、Spareデータ内の自由な領域を被らない領域へ変更 */
	.oobfree = {
		{.offset = 27,
		 .length =  1},
		{.offset = 31,
		 .length =  1},
		{.offset = 34,
		 .length = 30} }
#else
	.oobfree = {
		{.offset = 32,
		 .length = 32} }
#endif
};

/* flash bbt decriptors for large page devices */
static uint8_t bbt_pattern[] = {'B', 'b', 't', '0' };
static uint8_t mirror_pattern[] = {'1', 't', 'b', 'B' };

static struct nand_bbt_descr emma_bbt_main_descr_lp = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE
		| NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
#if 1 /* Mod for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/16 refs #34 @ vanhalen */
	/* BadBlockマーカー位置を spare[32],spare[33] に変えたので、Spareデータ内の自由な領域を被らない領域へ変更 */
	.offs =	34,
	.len = 4,
	.veroffs = 38, /* offs + len で 38 */
#else
	.offs =	32,
	.len = 4,
	.veroffs = 36,
#endif
	/* ↓ひとまずBBTパーティション領域のサイズ(block数は)defaultの4のままでこのdescripterをセットしておいて、NAND Flashの容量を検出後の emma_nand_scan_bbt() の中のBadBlockスキャンを行う前で、容量が512Mbitならば(最悪10blockの不良ブロックが発生(データシートより)するので14blockに、1Gbitならば24に変更する */
	.maxblocks = 4,
	.pattern = bbt_pattern
};

static struct nand_bbt_descr emma_bbt_mirror_descr_lp = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE
		| NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
#if 1 /* Mod for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/16 refs #34 @ vanhalen */
	.offs =	34,
	.len = 4,
	.veroffs = 38, /* offs + len で 38 */
#else
	.offs =	32,
	.len = 4,
	.veroffs = 36,
#endif
	/* ↓NAND Flashの容量を検出後の emma_nand_scan_bbt() の中で変更する */
	.maxblocks = 4,
	.pattern = mirror_pattern
};

/* #else */ /* CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
/* oob placement schemes for large and small page devices */
static struct nand_ecclayout emma_nand_oob_16 = {
	.eccbytes = 6,
	.eccpos = {1, 0, 2, 5, 4, 6},
#if 1 /* Mod for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/14 refs #34 @ vanhalen */
	/* BadBlockマーカー位置を spare[8] に変えたので、Spareデータ内の自由な領域を被らない領域へ変更 */
	.oobfree = {
		{.offset = 7,
		 . length = 1},
		{.offset = 9,
		 . length = 7} }
#else
	.oobfree = {
		{.offset = 8,
		 . length = 8} }
#endif
};

#if 1 /* Add for NAND Flash終端のBBTパーティション領域に不良blockが集中しても大丈夫にする by hiroshi.hasegawa@pixela.jp on 2012/03/09 refs #34 @ vanhalen */
static struct nand_bbt_descr emma_bbt_main_descr = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE
		| NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
#if 1 /* Mod for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/16 refs #34 @ vanhalen */
	/* BadBlockマーカー位置を spare[8] に変えたので、Spareデータ内の自由な領域を被らない領域へ変更 */
	.offs =	9,
	.len = 4,
	.veroffs = 13, /* offs + len で 13 */
#else
	.offs =	8,
	.len = 4,
	.veroffs = 12,
#endif
	/* ↓ひとまずBBTパーティション領域のサイズ(block数は)defaultの4のままでこのdescripterをセットしておいて、NAND Flashの容量を検出後の emma_nand_scan_bbt() の中のBadBlockスキャンを行う前で、容量が512Mbitならば(最悪80blockの不良ブロックが発生(データシートより)するので)84blockに、1Gbitならば(最悪160blockの不良ブロックが発生(データシートより)するので)164に変更する */
	.maxblocks = 4,
	.pattern = bbt_pattern
};

static struct nand_bbt_descr emma_bbt_mirror_descr = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE
		| NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
#if 1 /* Mod for Spareデータ内のBadBlockマーカー位置を変更 by hiroshi.hasegawa@pixela.jp on 2012/03/16 refs #34 @ vanhalen */
	.offs =	9,
	.len = 4,
	.veroffs = 13, /* offs + len で 13 */
#else
	.offs =	8,
	.len = 4,
	.veroffs = 12,
#endif
	/* ↓NAND Flashの容量を検出後の emma_nand_scan_bbt() の中で変更する */
	.maxblocks = 4,
	.pattern = mirror_pattern
};
#endif

/* #endif */ /* CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */

static void wait_ready_busy_bit_on(void)
{
	/* ROM Status Register: READY/BUSY bit */
	while ((readl((void *)NAND_ROM_STATUS) & NAND_ROM_STATUS_BUSY) == 0)
		;
}

static void wait_command_enable_bit_off(void)
{
	/* Nand ROM control Register: COMMAND_ENABLE bit */
	while ((readl((void *)NAND_ROM_COMMAND) & 0x00000001) == 1)
		;
}

static void wait_status_ok(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;

	chip->cmdfunc(mtd, NAND_CMD_STATUS, -1, -1);
	while ((chip->read_byte(mtd) & NAND_STATUS_READY) == 0)
		;
}

static void emma_nand_select_chip(struct mtd_info *mtd, int chip)
{
	if (chip == -1)
		writel(0x00000000, (void *)NAND_ROM_COMMAND);
}

static void emma_nand_write_buf(struct mtd_info *mtd,
				const uint8_t *buf,
				int len)
{
	int i;
	void __iomem *adr = ((struct nand_chip *)mtd->priv)->IO_ADDR_W;

	if ((len & 0x0f) == 0 && ((unsigned int)buf & 3) == 0) {
		while (len) {
			unsigned int a, b, c, d;
			a = *(unsigned int *)(buf + 0);
			b = *(unsigned int *)(buf + 4);
			c = *(unsigned int *)(buf + 8);
			d = *(unsigned int *)(buf + 12);
			writel(a, adr);
			writel(b, adr);
			writel(c, adr);
			writel(d, adr);
			buf += 16;
			len -= 16;
		}
	}  else  {
		for (i = 0; i < len; i++)
			writeb(buf[i], adr);
	}
}

static void emma_nand_read_buf(struct mtd_info *mtd, uint8_t *buf, int len)
{
	int i;
	void __iomem *adr = ((struct nand_chip *)mtd->priv)->IO_ADDR_R;

	if ((len & 0x0f) == 0 && ((unsigned int)buf & 3) == 0) {
		while (len) {
			unsigned int a, b, c, d;
			a = readl(adr);
			b = readl(adr);
			c = readl(adr);
			d = readl(adr);
			*(unsigned int *)(buf + 0) = a;
			*(unsigned int *)(buf + 4) = b;
			*(unsigned int *)(buf + 8) = c;
			*(unsigned int *)(buf + 12) = d;
			buf += 16;
			len -= 16;
		}
	}  else  {
		for (i = 0; i < len; i++)
			buf[i] = readb(adr);
	}
}

static void emma_nand_cmd_ctrl(struct mtd_info *mtd,
				  int dat,
				  unsigned int ctrl)
{
}

static int emma_nand_dev_ready(struct mtd_info *mtd)
{
/*	wait_ready_busy_bit_on(); 本関数は nand_base.c の nand_wait() 内でR/Bピン状態監視のpolling毎にCallされており、本関数が1(ready)を返すまでwhileループ待ちをするようになっているので、本関数内でココでbusy終了までspin待ちをする必要はありません。というか、本関数が0(busy)を返した際にwatchdogのリセットやフェールセーフ(タイムアウト)処理を nand_wait() 内でちゃんと行うようにしてあるので、ココでbusy終了までspin待ちしてはダメ！ */
	return readl((void *)NAND_ROM_STATUS) & NAND_ROM_STATUS_BUSY ? 1 : 0;
}

/**
 * emma_nand_check_wp - [GENERIC] check if the chip is write protected
 * @mtd:	MTD device structure
 * Check, if the device is write protected
 *
 * The function expects, that the device is already selected
 */
static int emma_nand_check_wp(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	/* Check the WP bit */
	chip->cmdfunc(mtd, NAND_CMD_STATUS, -1, -1);
	return (chip->read_byte(mtd) & NAND_STATUS_WP) ? 0 : 1;
}

/*
 * Send command to NAND device
 */
/* #if CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
static void emma_nand_cmdfunc_lp(struct mtd_info *mtd,
				    unsigned command,
				    int column,
				    int page_addr)
{
	struct nand_chip *chip = mtd->priv;

	/* Emulate NAND_CMD_READOOB */
	if (command == NAND_CMD_READOOB) {
		column += mtd->writesize;
		command = NAND_CMD_READ0;
	}

	switch (command) {
	case NAND_CMD_SEQIN:
	/*	wait_status_ok(mtd); ↓の wait_ready_busy_bit_on() がbusy終了までの待ち処理なので不要 */
		wait_command_enable_bit_off();
		wait_ready_busy_bit_on();

		writel(((page_addr & 0x0000FFFF) << 16) | (column & 0x0000FFFF),
		       (void *)NAND_ROM_ADDR1);
		writel((page_addr & 0x00FF0000) >> 16, (void *)NAND_ROM_ADDR2);

		if (chip->chipsize <= (128 << 20)) { /* 128MB */
			writel(0x3017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x4017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		break;

	case NAND_CMD_PAGEPROG:
	case NAND_CMD_CACHEDPROG:
		wait_command_enable_bit_off();
		writel(0x0005 | (command << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
	/*	wait_status_ok(mtd); なぜココでNAND_CMD_STATUS(Read Statusコマンド)発行＆Status Read＆busy終了まで待つ！？ そもそもココでは↑のNAND_CMD_PAGEPROG(Programコマンド)の発行が完了していないので、おそらくココでのNAND_CMD_STATUS(Read Statusコマンド)発行は無視されています */
		wait_command_enable_bit_off();
	/*	wait_ready_busy_bit_on(); なぜココでbusy終了まで待つ！？ ココではまだR/BピンがLow(busy)になってない可能性もあり、意味がありません */
		break;

	case NAND_CMD_READ0:
#if 1 /* Mod for NAND FlashのReady/Busyステータスを(オーバーヘッドの大きい)ReadStatusコマンドでなく、R/Bピンのレベルで取得する by hiroshi.hasegawa@pixela.jp */
		wait_ready_busy_bit_on();
#else
		wait_status_ok(mtd);
#endif
		writel(((page_addr & 0x0000FFFF) << 16) | (column & 0x0000FFFF),
		       (void *)NAND_ROM_ADDR1);
		writel((page_addr & 0x00FF0000) >> 16, (void *)NAND_ROM_ADDR2);

#if 1 /* Mod for R/BピンがHigh(ready)になってからpageデータの読み出しを開始するまでのtRR(20nsec)待ちを行うために、READコマンド発行～R/BピンのHigh(ready)までの待ち処理をマニュアルで行う by hiroshi.hasegawa@pixela.jp */
		/* コマンド発行を(FlashROMメモリ空間へのReadアクセス発生後ではなく)すぐに行うように COMMAND_START[1]
		 * を 0b へ、R/BピンのHigh(ready)待ちをマニュアルで行うので READ_MODE[6] を 0b(=Don't wait) へ */
		if (chip->chipsize <= (128 << 20)) { /* 128MB */
			writel(0x301D | (command << NAND_CMD0_SHIFT)|
					(NAND_CMD_READSTART << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x401D | (command << NAND_CMD0_SHIFT)|
					(NAND_CMD_READSTART << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		wait_command_enable_bit_off(); /* コマンド発行完了待ち */
		ndelay(100); /* R/Bピンがbusy(Low)に落ち込むまで、tWB(100nsec)待たなければダメ！ */
		nand_wait_ready(mtd); /* R/BピンがHigh(ready)になるまで待ち */
		ndelay(20); /* tRR(20nsec)待ち */
#else
		writel(0x405F | (command << NAND_CMD0_SHIFT)|
				(NAND_CMD_READSTART << NAND_CMD1_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);

		wait_ready_busy_bit_on();
#endif
		break;

	case NAND_CMD_ERASE1:
		wait_status_ok(mtd);
		writel(page_addr, (void *)NAND_ROM_ADDR1);

		if (chip->chipsize <= (128 << 20)) { /* 128MB */
			writel(0x101D | (NAND_CMD_ERASE1 << NAND_CMD0_SHIFT) |
					(NAND_CMD_ERASE2 << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x201D | (NAND_CMD_ERASE1 << NAND_CMD0_SHIFT) |
					(NAND_CMD_ERASE2 << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}

	/*	wait_status_ok(mtd); なぜココでNAND_CMD_STATUS(Read Statusコマンド)発行＆Status Read＆busy終了まで待つ！？ そもそもココでは↑のNAND_CMD_ERASE1(Programコマンド)の発行が完了していないので、おそらくココでのNAND_CMD_STATUS(Read Statusコマンド)発行は無視されています */
		wait_command_enable_bit_off();
	/*	wait_ready_busy_bit_on(); なぜココでbusy終了まで待つ！？ ココではまだR/BピンがLow(busy)になってない可能性もあり、意味がありません */
		break;

	case NAND_CMD_ERASE2:
		/* marge "NAND_CMD_ERASE1:" */
		break;

	case NAND_CMD_STATUS:
		writel(0x0007 | NAND_CMD_STATUS << NAND_CMD0_SHIFT |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
		break;

	case NAND_CMD_READID:
		wait_status_ok(mtd);
		writel(0x00000000, (void *)NAND_ROM_ADDR1);
		writel(0x0017 | NAND_CMD_READID << NAND_CMD0_SHIFT |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
		break;

	case NAND_CMD_RESET:
		writel(0x0005 | NAND_CMD_RESET << NAND_CMD0_SHIFT |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
		wait_command_enable_bit_off(); /* コマンド発行完了を待たなきゃダメ！ */
		ndelay(100); /* R/Bピンがbusy(Low)に落ち込むまで、tWB(100nsec)待たなければダメ！ */
		wait_ready_busy_bit_on();
		writel(0x00000000, (void *)NAND_ROM_COMMAND);
		break;
	}
}
/* #else */ /* CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
static void emma_nand_cmdfunc(struct mtd_info *mtd,
				 unsigned command,
				 int column,
				 int page_addr)
{
	struct nand_chip *chip = mtd->priv;
	int readcmd;

	switch (command) {
	case NAND_CMD_SEQIN:

	/*	wait_status_ok(mtd); ↓の wait_ready_busy_bit_on() がbusy終了までの待ち処理なので不要 */
		wait_command_enable_bit_off();
		wait_ready_busy_bit_on();

		if (column >= mtd->writesize) {
			/* OOB area */
			column -= mtd->writesize;
			readcmd = NAND_CMD_READOOB;
		} else if (column < 256) {
			/* First 256 bytes --> READ0 */
			readcmd = NAND_CMD_READ0;
		} else {
			column -= 256;
			readcmd = NAND_CMD_READ1;
		}

		writel(0x0005 | (readcmd << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);

		wait_command_enable_bit_off();

		writel(((page_addr & 0x00FFFFFF) << 8) | (column & 0x000000FF),
		       (void *)NAND_ROM_ADDR1);

		if (chip->chipsize <= (32 << 20)) { /* 32MB */
			writel(0x2017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x3017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		break;

	case NAND_CMD_PAGEPROG:
		wait_command_enable_bit_off();
		writel(0x0005 | (NAND_CMD_PAGEPROG << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
	/*	wait_status_ok(mtd); なぜココでNAND_CMD_STATUS(Read Statusコマンド)発行＆Status Read＆busy終了まで待つ！？ そもそもココでは↑のNAND_CMD_PAGEPROG(Programコマンド)の発行が完了していないので、おそらくココでのNAND_CMD_STATUS(Read Statusコマンド)発行は無視されています */
		wait_command_enable_bit_off();
	/*	wait_ready_busy_bit_on(); なぜココでbusy終了まで待つ！？ ココではまだR/BピンがLow(busy)になってない可能性もあり、意味がありません */
		break;

	case NAND_CMD_READ0:
	case NAND_CMD_READ1:
	case NAND_CMD_READOOB:
#if 1 /* Mod for NAND FlashのReady/Busyステータスを(オーバーヘッドの大きい)ReadStatusコマンドでなく、R/Bピンのレベルで取得する by hiroshi.hasegawa@pixela.jp */
		wait_ready_busy_bit_on();
#else
		wait_status_ok(mtd);
#endif
		writel(((page_addr & 0x00FFFFFF) << 8) | (column & 0x000000FF),
		       (void *)NAND_ROM_ADDR1);

#if 1 /* Mod for R/BピンがHigh(ready)になってからpageデータの読み出しを開始するまでのtRR(20nsec)待ちを行うために、READコマンド発行～R/BピンのHigh(ready)までの待ち処理をマニュアルで行う by hiroshi.hasegawa@pixela.jp */
		/* コマンド発行を(FlashROMメモリ空間へのReadアクセス発生後ではなく)すぐに行うように COMMAND_START[1]
		 * を 0b へ、R/BピンのHigh(ready)待ちをマニュアルで行うので READ_MODE[6] を 0b(=Don't wait) へ */
		if (chip->chipsize <= (32 << 20)) {	/* 32M */
			writel(0x2015 | (command << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x3015 | (command << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		wait_command_enable_bit_off(); /* コマンド発行完了待ち */
		ndelay(100); /* R/Bピンがbusy(Low)に落ち込むまで、tWB(100nsec)待たなければダメ！ */
		nand_wait_ready(mtd); /* R/BピンがHigh(ready)になるまで待ち */
		ndelay(20); /* tRR(20nsec)待ち */
#else
		if (chip->chipsize < 0x04000000)	/* 64M */
			writel(0x2057 | (command << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		else
			writel(0x3057 | (command << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		wait_ready_busy_bit_on();
#endif
		break;

	case NAND_CMD_ERASE1:
		wait_status_ok(mtd);
		writel(page_addr, (void *)NAND_ROM_ADDR1);
		if (chip->chipsize <= (32 << 20)) {	/* 32M */
			writel(0x101D | (NAND_CMD_ERASE1 << NAND_CMD0_SHIFT) |
					(NAND_CMD_ERASE2 << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x201D | (NAND_CMD_ERASE1 << NAND_CMD0_SHIFT) |
					(NAND_CMD_ERASE2 << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
	/*	wait_status_ok(mtd); なぜココでNAND_CMD_STATUS(Read Statusコマンド)発行＆Status Read＆busy終了まで待つ！？ そもそもココでは↑のNAND_CMD_ERASE1/2(Eraseコマンド)の発行が完了していないので、おそらくココでのNAND_CMD_STATUS(Read Statusコマンド)発行は無視されています */
		wait_command_enable_bit_off();
	/*	wait_ready_busy_bit_on(); なぜココでbusy終了まで待つ！？ ココではまだR/BピンがLow(busy)になってない可能性もあり、意味がありません */
		break;

	case NAND_CMD_ERASE2:
		/* marge "NAND_CMD_ERASE1:" */
		break;

	case NAND_CMD_STATUS:
		writel(0x0007 | (NAND_CMD_STATUS << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
		break;

	case NAND_CMD_READID:
		wait_status_ok(mtd);
		writel(0x00000000, (void *)NAND_ROM_ADDR1);
		writel(0x0017 | (NAND_CMD_READID << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
		break;

	case NAND_CMD_RESET:
		writel(0x0005 | (NAND_CMD_RESET << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);
		wait_command_enable_bit_off(); /* コマンド発行完了を待たなきゃダメ！ */
		ndelay(100); /* R/Bピンがbusy(Low)に落ち込むまで、tWB(100nsec)待たなければダメ！ */
		wait_ready_busy_bit_on();
		writel(0x00000000, (void *)NAND_ROM_COMMAND);
		break;
	}
}
/* #endif */ /* CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */

/* #if CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
static void emma_nand_seq_cmd_lp(
	struct mtd_info *mtd,
	unsigned command,
	int page_addr,
	int size_bit)
{
	struct nand_chip *chip = mtd->priv;

	writel(((page_addr & 0x0000FFFF) << 16), (void *)NAND_ROM_ADDR1);
	writel((page_addr & 0x00FF0000) >> 16, (void *)NAND_ROM_ADDR2);

	switch (command) {
	case NAND_SEQ_READ:
		writel(0x00000101 | size_bit, (void *)NAND_ROM_CONTROL3);
		if (chip->chipsize <= (128 << 20)) { /* 128MB */
			writel(0x305F | (command << NAND_CMD0_SHIFT)|
					(NAND_CMD_READSTART << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x405F | (command << NAND_CMD0_SHIFT)|
					(NAND_CMD_READSTART << NAND_CMD1_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		break;

	case NAND_SEQ_WRITE:
		writel(0x01700101 | size_bit, (void *)NAND_ROM_CONTROL3);
		if (chip->chipsize <= (128 << 20)) { /* 128MB */
			writel(0x3017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT)
				      | (NAND_CMD_PAGEPROG << NAND_CMD1_SHIFT)
				      | (NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x4017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT)
				      | (NAND_CMD_PAGEPROG << NAND_CMD1_SHIFT)
				      | (NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		break;
	}
}
/* #else */ /* CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
static void emma_nand_seq_cmd(
	struct mtd_info *mtd,
	unsigned command,
	int page_addr,
	int size_bit)
{
	struct nand_chip *chip = mtd->priv;

	switch (command) {
	case NAND_SEQ_READ:
		writel(((page_addr & 0x00FFFFFF) << 8),
		       (void *)NAND_ROM_ADDR1);

		writel(0x00000001 | size_bit, (void *)NAND_ROM_CONTROL3);

		if (chip->chipsize <= (32 << 20)) { /* 32MB */
			writel(0x2057 | (NAND_CMD_READ0 << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x3057 | (NAND_CMD_READ0 << NAND_CMD0_SHIFT) |
					(NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		break;

	case NAND_SEQ_WRITE:
		writel(0x0005 | (NAND_CMD_READ0 << NAND_CMD0_SHIFT) |
				(NAND_CS_BIT << NAND_CS_SHIFT),
		       (void *)NAND_ROM_COMMAND);

		wait_command_enable_bit_off();

		writel(((page_addr & 0x00FFFFFF) << 8),
		       (void *)NAND_ROM_ADDR1);

		writel(0x01700001 | size_bit, (void *)NAND_ROM_CONTROL3);

		if (chip->chipsize <= (32 << 20)) { /* 32MB */
			writel(0x2017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT)
				      | (NAND_CMD_PAGEPROG << NAND_CMD1_SHIFT)
				      | (NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		} else {
			writel(0x3017 | (NAND_CMD_SEQIN << NAND_CMD0_SHIFT)
				      | (NAND_CMD_PAGEPROG << NAND_CMD1_SHIFT)
				      | (NAND_CS_BIT << NAND_CS_SHIFT),
			       (void *)NAND_ROM_COMMAND);
		}
		break;
	}
}
/* #endif */ /* CONFIG_SYS_NAND_LARGE_PAGE */ /* Del for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */

/**
 * emma_nand_get_device - [GENERIC] Get chip for selected access
 * @chip:	the nand chip descriptor
 * @mtd:	MTD device structure
 * @new_state:	the state which is requested
 *
 * Get the device and lock it for exclusive access
 */
static int
emma_nand_get_device(
	struct nand_chip *chip,
	struct mtd_info *mtd,
	int new_state)
{
	chip->state = new_state;
	return 0;
}

/**
 * emma_nand_release_device - [GENERIC] release chip
 * @mtd:	MTD device structure
 *
 * Deselect, release chip lock and wake up anyone waiting on the device
 */
static void emma_nand_release_device(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;

	chip->select_chip(mtd, -1);	/* De-select the NAND device */
}

void emma_nand_correct_data(struct mtd_info *mtd,
			    uint8_t *buf_poi,
			    int page_num)
{
	int j, k;
	int ecc_res;
	int ecc_byte, ecc_bit;
	int reg_num;
	struct nand_chip *chip = mtd->priv;
	int ssr = readl((void *)NAND_ROM_SEQ_STATUS);

	if ((ssr & 0x02000000) || (ssr & 0x04000000)) { /* 1 or 2 bit error */
		reg_num = page_num * mtd->writesize / (chip->ecc.size * 2);

		for (j = 0; j < reg_num; j++) {
			ecc_res = readl((void *)NAND_ROM_ECC_RESULT + (j * 4));

			for (k = 0; k < 2; k++) {
				if (ecc_res & 0x00000002) {
					/* 1 bit error */
					ecc_byte = (j * chip->ecc.size * 2)
						   + (k * chip->ecc.size)
						   + ((ecc_res & 0x0000FF00)
						      >> 0x08);
					ecc_bit = 0x01
						  << ((ecc_res & 0x000000E0)
						      >> 0x05);

					/* correct */
					buf_poi[ecc_byte] ^=  ecc_bit;
					mtd->ecc_stats.corrected++;

				} else if (ecc_res & 0x00000004) {
					/* 2 bit error */
					mtd->ecc_stats.failed++;
				}

				ecc_res = ecc_res >> 0x10;
			}
		}
	}

	writel(0x00000001, (void *)NAND_ROM_CONTROL6);	/* ecc circuit reset */
}

#if 0 /* Del by hiroshi.hasegawa@pixela.jp */
/**
 * emma_nand_do_read_ops - [Internal] Read data with ECC
 *
 * @mtd:	MTD device structure
 * @from:	offset to read from
 * @ops:	oob ops structure
 *
 * Internal function. Called with chip held.
 */
static int emma_nand_do_read_ops(
	struct mtd_info *mtd,
	loff_t from,
	struct mtd_oob_ops *ops)
{
	int chipnr, page, realpage, col, bytes, aligned;
	struct nand_chip *chip = mtd->priv;
	struct mtd_ecc_stats stats;
	int ret = 0;
	uint32_t readlen = ops->len;
	uint8_t *bufpoi, *buf;
	int page_num;
	int seq_bytes;
	int page_tmp;
	int size_bit;
	int carry;

	stats = mtd->ecc_stats;

	chipnr = (int)(from >> chip->chip_shift);
	chip->select_chip(mtd, chipnr);

	realpage = (int)(from >> chip->page_shift);
	page = realpage & chip->pagemask;

	col = (int)(from & (mtd->writesize - 1));
	buf = ops->datbuf;

	while (1) {
		page_num = 1;
		bytes = min(mtd->writesize - col, readlen);
		aligned = (bytes == mtd->writesize);

		/* Is the current page in the buffer ? */
		if (realpage != chip->pagebuf) {
			bufpoi = aligned ? buf : chip->buffers->databuf;

			if (aligned) {
				page_tmp
				= min(readlen,
				      (uint32_t)NAND_SEQ_MAX_SIZE)
				  / mtd->writesize;

				if (page_tmp < 2) {
					page_num = 1;
					size_bit = 0x00;
				} else if (page_tmp < 4) {
					page_num = 2;
					size_bit = 0x10;
				} else if (page_tmp < 8) {
					page_num = 4;
					size_bit = 0x20;
				} else if (page_tmp < 16) {
					page_num = 8;
					size_bit = 0x30;
				} else if (page_tmp < 32) {
					page_num = 16;
					size_bit = 0x40;
				} else if (page_tmp < 64) {
					page_num = 32;
					size_bit = 0x50;
				} else {
					page_num = 64;
					size_bit = 0x60;
				}

				carry = ((page + page_num) & 0x000f00)
					- (page & 0x000f00);
				if (carry != 0) {
					page_num = 1;
					size_bit = 0x00;
				}

				seq_bytes = page_num * mtd->writesize;
				bytes = seq_bytes;
			} else {
				seq_bytes = mtd->writesize;
				page_num = 1;
				size_bit = 0x00;
			}

			emma_nand_info.seq_cmdfunc(mtd, NAND_SEQ_READ,
						   page, size_bit);
			chip->read_buf(mtd, bufpoi, seq_bytes);

			while ((readl((void *)NAND_ROM_INTERRUPT)
			       & 0x00000100) == 0)
				;/* wait seq end */

			writel(0x00000000, (void *)NAND_ROM_INTERRUPT);

			emma_nand_correct_data(mtd, bufpoi, page_num);

			/* Transfer not aligned data */
			if (!aligned) {
				chip->pagebuf = realpage;
				memcpy(buf,
				       chip->buffers->databuf + col,
				       bytes);
			}

			buf += bytes;
			if (!(chip->options & NAND_NO_READRDY)) {
				/*
				 * Apply delay or wait for ready/busy
				 * pin. Do this before the AUTOINCR
				 * check, so no problems arise
				 * if a chip which does auto
				 * increment is marked as NOAUTOINCR
				 * by the board driver.
				 */
				if (!chip->dev_ready)
					udelay(chip->chip_delay);
				else
					nand_wait_ready(mtd);
			}
		} else {
			memcpy(buf, chip->buffers->databuf + col,
			       bytes);
			buf += bytes;
		}
		readlen -= bytes;

		if (!readlen)
			break;

		/* For subsequent reads align to page boundary. */
		col = 0;
		/* Increment page address */
		realpage += page_num;

		page = realpage & chip->pagemask;
		/* Check, if we cross a chip boundary */
		if (!page) {
			chipnr++;
			chip->select_chip(mtd, -1);
			chip->select_chip(mtd, chipnr);
		}
	}
	writel(0x00000000, (void *)NAND_ROM_COMMAND);

	ops->retlen = ops->len - (size_t) readlen;

	if (ret)
		return ret;

	if (mtd->ecc_stats.failed - stats.failed)
		return -EBADMSG;

	return  mtd->ecc_stats.corrected - stats.corrected ? -EUCLEAN : 0;
}

/**
 * emma_nand_read - [MTD Interface] MTD compability function for
 *                  nand_do_read_ecc
 * @mtd:	MTD device structure
 * @from:	offset to read from
 * @len:	number of bytes to read
 * @retlen:	pointer to variable to store the number of read bytes
 * @buf:	the databuffer to put data
 *
 * Get hold of the chip and call nand_do_read
 */
static int emma_nand_read(
	struct mtd_info *mtd,
	loff_t from,
	size_t len,
	size_t *retlen,
	uint8_t *buf)
{
	struct nand_chip *chip = mtd->priv;
	int ret;

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size)
		return -EINVAL;
	if (!len)
		return 0;

	emma_nand_get_device(chip, mtd, FL_READING);

	chip->ops.len = len;
	chip->ops.datbuf = buf;
	chip->ops.oobbuf = NULL;

	ret = emma_nand_do_read_ops(mtd, from, &chip->ops);

	*retlen = chip->ops.retlen;

	emma_nand_release_device(mtd);

	return ret;
}
#endif /* Del by hiroshi.hasegawa@pixela.jp */

#define NOTALIGNED(x)	((x & (chip->subpagesize - 1)) != 0)

/**
 * emma_nand_do_write_ops - [Internal] NAND write with ECC
 * @mtd:	MTD device structure
 * @to:		offset to write to
 * @ops:	oob operations description structure
 *
 * NAND write with ECC
 */
static int emma_nand_do_write_ops(struct mtd_info *mtd, loff_t to,
			     struct mtd_oob_ops *ops)
{
	int chipnr, realpage, page, blockmask, column;
	struct nand_chip *chip = mtd->priv;
	uint32_t writelen = ops->len;
	uint8_t *buf = ops->datbuf;
	int ret = 0;
	int subpage;
	int bytes;
	int page_num;
	int page_tmp;
	int size_bit;
	int ssr, status;
	int carry;

	ops->retlen = 0;
	if (!writelen)
		return 0;

	/* reject writes, which are not page aligned */
	if (NOTALIGNED(to) || NOTALIGNED(ops->len)) {
		printk(KERN_NOTICE "emma_nand_write: "
		       "Attempt to write not page aligned data\n");
		return -EINVAL;
	}

	column = to & (mtd->writesize - 1);
	subpage = column || (writelen & (mtd->writesize - 1));

	if (subpage)
		return -EINVAL;

	chipnr = (int)(to >> chip->chip_shift);
	chip->select_chip(mtd, chipnr);

	/* Check, if it is write protected */
	if (emma_nand_check_wp(mtd))
		return -EIO;

	realpage = (int)(to >> chip->page_shift);
	page = realpage & chip->pagemask;
	blockmask = (1 << (chip->phys_erase_shift - chip->page_shift)) - 1;

	/* Invalidate the page cache, when we write to the cached page */
	if (to <= (chip->pagebuf << chip->page_shift) &&
	    (chip->pagebuf << chip->page_shift) < (to + ops->len))
		chip->pagebuf = -1;

	while (1) {
		uint8_t *wbuf = buf;

		bytes = mtd->writesize;

		/* Partial page write ? */
		if (unlikely(column
			     || writelen < (mtd->writesize - 1))) {
			bytes = min_t(int, bytes - column,
				      (int) writelen);
			chip->pagebuf = -1;
			memset(chip->buffers->databuf, 0xff,
			       mtd->writesize);
			memcpy(&chip->buffers->databuf[column],
			       buf, bytes);
			wbuf = chip->buffers->databuf;
			page_num = 1;
			size_bit = 0x00;
		} else {
			page_tmp = min(writelen,
				       (uint32_t)NAND_SEQ_MAX_SIZE)
				   / mtd->writesize;
			if (page_tmp < 2) {
				page_num = 1;
				size_bit = 0x00;
			} else if (page_tmp < 4) {
				page_num = 2;
				size_bit = 0x10;
			} else if (page_tmp < 8) {
				page_num = 4;
				size_bit = 0x20;
			} else if (page_tmp < 16) {
				page_num = 8;
				size_bit = 0x30;
			} else if (page_tmp < 32) {
				page_num = 16;
				size_bit = 0x40;
			} else if (page_tmp < 64) {
				page_num = 32;
				size_bit = 0x50;
			} else {
				page_num = 64;
				size_bit = 0x60;
			}

			carry = ((page + page_num) & 0x000f00)
				- (page & 0x000f00);
			if (carry != 0) {
				page_num = 1;
				size_bit = 0x00;
			}

			bytes = page_num * mtd->writesize;
		}

		emma_nand_info.seq_cmdfunc(mtd, NAND_SEQ_WRITE, page,
					   size_bit);
		chip->write_buf(mtd, wbuf, bytes);
		while ((readl((void *)NAND_ROM_INTERRUPT)
		       & 0x00000100) == 0)
			;/* wait seq end */

		writel(0x00000000, (void *)NAND_ROM_INTERRUPT);

		ssr = readl((void *)NAND_ROM_SEQ_STATUS);

		if (ssr & 0x80000000) {	/* program error */
			/* program status reset */
			writel(0x00000002, (void *)NAND_ROM_CONTROL6);
			DPRINTK(2, "emma_nand_write_ecc: "
				"Failed ECC read, page 0x%08x\n",
				page);
			ret = -EIO;
			break;
		}

		/* program status reset */
		writel(0x00000002, (void *)NAND_ROM_CONTROL6);

		status = chip->waitfunc(mtd, chip);

		/*
		 * See if operation failed and additional status
		 * hecks are available
		 */
		if ((status & NAND_STATUS_FAIL) && (chip->errstat))
			status = chip->errstat(mtd, chip, FL_WRITING,
					       status, page);

		if (status & NAND_STATUS_FAIL) {
			ret =  -EIO;
			break;
		}

		writelen -= bytes;
		if (!writelen)
			break;

		column = 0;
		buf += bytes;
		realpage += page_num;

		page = realpage & chip->pagemask;
		/* Check, if we cross a chip boundary */
		if (!page) {
			chipnr++;
			chip->select_chip(mtd, -1);
			chip->select_chip(mtd, chipnr);
		}
	}

	writel(0x00000000, (void *)NAND_ROM_COMMAND);

	ops->retlen = ops->len - writelen;
	return ret;
}

/**
 * emma_nand_write - [MTD Interface] NAND write with ECC
 * @mtd:	MTD device structure
 * @to:		offset to write to
 * @len:	number of bytes to write
 * @retlen:	pointer to variable to store the number of written bytes
 * @buf:	the data to write
 *
 * NAND write with ECC
 */
static int emma_nand_write(
	struct mtd_info *mtd,
	loff_t to, size_t len,
	size_t *retlen,
	const uint8_t *buf)
{
	struct nand_chip *chip = mtd->priv;
	int ret;

	/* Do not allow reads past end of device */
	if ((to + len) > mtd->size)
		return -EINVAL;
	if (!len)
		return 0;

	emma_nand_get_device(chip, mtd, FL_WRITING);

	chip->ops.len = len;
	chip->ops.datbuf = (uint8_t *)buf;
	chip->ops.oobbuf = NULL;

	ret = emma_nand_do_write_ops(mtd, to, &chip->ops);

	*retlen = chip->ops.retlen;

	emma_nand_release_device(mtd);

	return ret;
}

#if 1 /* Add for NAND Flash終端のBBTパーティション領域に不良blockが集中しても大丈夫にする by hiroshi.hasegawa@pixela.jp on 2012/03/14 refs #34 @ vanhalen */
static int emma_nand_scan_bbt(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	uint32_t bad_block_num_max = 0;
	
	if (chip->bbt_td == &emma_bbt_main_descr && chip->bbt_md == &emma_bbt_mirror_descr) {
		/* SmallPageの場合 */
		if ((128LU*1024*1024) <= mtd->size) { /* NAND Flashの容量が1Gbit(128Mbyte)の場合 */
			bad_block_num_max = 164;          /* 不良blockが最悪160個発生する            */
		} else {                    /* NAND Flashの容量が512bit(64Mbyte)の場合 */
			bad_block_num_max = 84; /* 不良blockが最悪80個発生する             */
		}
	} else if (chip->bbt_td == &emma_bbt_main_descr_lp && chip->bbt_md == &emma_bbt_mirror_descr_lp) {
		/* LargePageの場合 */
		if ((128LU*1024*1024) <= mtd->size) { /* NAND Flashの容量が1Gbit(128Mbyte)の場合 */
/* printk("<0>--------> change chip->bbt_td->maxblocks\n"); */
			bad_block_num_max = 24;           /* 不良blockが最悪20個発生                 */
		} else {                    /* NAND Flashの容量が512bit(64Mbyte)の場合 */
			bad_block_num_max = 14; /* 不良blockが最悪10個発生するので         */
		}
	}
	
	if (bad_block_num_max) {
/* printk("<0>--------> chip->bbt_td->maxblocks %u -> %u\n", chip->bbt_td->maxblocks, chip->bbt_td->maxblocks + bad_block_num_max); */
		chip->bbt_td->maxblocks = bad_block_num_max;
		chip->bbt_md->maxblocks = bad_block_num_max;
	}
	
	return nand_default_bbt(mtd);
}
#endif

void board_nand_init_tail(struct mtd_info *mtd)
{
/*	mtd->read = emma_nand_read; emma_nand_read() によるSequenceRead処理の方が速いのだが、R/BピンがHigh(ready)になってからpageデータの読み出しを開始するまでのtRR(20nsec)待ちを保障出来ないため、mtdのdefaultの nand_base.c 内の nand_read() でpage読み出しを行う */
	mtd->write = emma_nand_write;
}

int board_nand_init(struct nand_chip *chip)
{
	writel(0x00000000, (void *)NAND_ROM_INTERRUPT);
	writel(0x00000000, (void *)NAND_ROM_INTERRUPT_ENABLE);

	chip->write_buf = emma_nand_write_buf;
	chip->read_buf = emma_nand_read_buf;
	chip->cmd_ctrl = emma_nand_cmd_ctrl;
	chip->dev_ready = emma_nand_dev_ready;
	chip->select_chip = emma_nand_select_chip;
#if 1 /* Add for NAND Flash終端のBBTパーティション領域に不良blockが集中しても大丈夫にする by hiroshi.hasegawa@pixela.jp on 2012/03/09 refs #34 @ vanhalen */
	chip->scan_bbt = emma_nand_scan_bbt;
#endif

	chip->ecc.mode = NAND_ECC_SOFT;

	chip->options = NAND_USE_FLASH_BBT;

#if 1 /* Mod for NAND FlashのSmall/LargePage判別をStrapピンから動的に行う by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
	if (readl((void *)BHIF_STRAP_0) & (1LU << 21)) { /* [21:20] BOOT_NAND_SEL が 00b,01b ならば SmallPage、10b,11b ならば LargePage */
		/* large page */
		chip->cmdfunc = emma_nand_cmdfunc_lp;
		chip->ecc.layout = &emma_nand_oob_64;
		chip->bbt_td = &emma_bbt_main_descr_lp;
		chip->bbt_md = &emma_bbt_mirror_descr_lp;
		chip->badblock_pattern = &emma_largepage_flashbased;
		emma_nand_info.seq_cmdfunc = emma_nand_seq_cmd_lp;
	} else {
		/* small page */
		chip->cmdfunc = emma_nand_cmdfunc;
		chip->ecc.layout = &emma_nand_oob_16;
		chip->bbt_td = &emma_bbt_main_descr;
		chip->bbt_md = &emma_bbt_mirror_descr;
		chip->badblock_pattern = &emma_smallpage_flashbased;
		emma_nand_info.seq_cmdfunc = emma_nand_seq_cmd;
	}
#else  /* Mod by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */
#if CONFIG_SYS_NAND_LARGE_PAGE /* large page */
	chip->cmdfunc = emma_nand_cmdfunc_lp;
	chip->ecc.layout = &emma_nand_oob_64;
	chip->bbt_td = &emma_bbt_main_descr_lp;
	chip->bbt_md = &emma_bbt_mirror_descr_lp;
	emma_nand_info.seq_cmdfunc = emma_nand_seq_cmd_lp;
#else	/* small page */
	chip->cmdfunc = emma_nand_cmdfunc;
	chip->ecc.layout = &emma_nand_oob_16;
	emma_nand_info.seq_cmdfunc = emma_nand_seq_cmd;
#endif
#endif /* Mod by hiroshi.hasegawa@pixela.jp on 2012/03/06 refs #34 @ vanhalen */

#if CONFIG_SYS_NAND_CACHEPRG_SUPPORT
	chip->options |= NAND_CACHEPRG;
#endif
	return 0;
}

