/* NEC candy ethernet driver
 *
 * (C) Copyright 2005 Hiroyasu Nakatsuka
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 *    REVISION HISTORY:
 *    ================
 *
 *    v0.5       7-7-2005      nakatuka        Suport eth0 only
 *
 */

#include <config.h>

#ifdef CONFIG_NEC_CANDY

#include <common.h>
#include <malloc.h>
#include <net.h>
#include <miiphy.h>
#include <command.h>
#include <nec_candy_pd.h>
#include <asm/io.h>
#include <asm/cacheops.h>
#include <asm/arch-emma/emma.h>
#include "nec_candy.h"

/***********************************************************************
 * debug
 ***********************************************************************/

/* local debug macro */
#undef DEBUG_NEC_CANDY

#undef debug
#ifdef DEBUG_NEC_CANDY
#define debug(fmt, args...)	printf(fmt, ##args)
#else
#define debug(fmt, args...)
#endif

static volatile struct candy_desc rx_descs[RX_RING_SIZE + 1];
static volatile struct candy_desc tx_descs[TX_RING_SIZE + 1];

#define MAX_WAIT 1000

static int candy_rmii = 1;	/* 0:MII 1:RMII */

/***********************************************************************
 * helpers
 ***********************************************************************/
#define	candy_in(x)	readl((void *) &x)
#define	candy_out(x, y)	writel((unsigned long)y, (void *)&x)

#define	candy_set_bits(x, mask)  candy_out(x, candy_in(x) | mask)
#define	candy_clear_bits(x, mask)  candy_out(x, candy_in(x) & ~mask)

#define	candy_set_macc1_bits(x, mask) \
		candy_out(x, (candy_in(x) | mask) & MACC1_RESERVED)
#define	candy_clear_macc1_bits(x, mask) \
		candy_out(x, (candy_in(x) & ~mask) & MACC1_RESERVED)

#define	candy_set_macc2_bits(x, mask) \
		candy_out(x, (candy_in(x) | mask) & MACC2_RESERVED)
#define	candy_clear_macc2_bits(x, mask) \
		candy_out(x, (candy_in(x) & ~mask) & MACC2_RESERVED)

#define MEMSYNC(x)							\
	do {								\
		volatile unsigned long *p = (unsigned long *)x;		\
		(void)(*p);						\
		__asm__ __volatile__("sync");				\
	} while (0)

/***********************************************************************
 * low-level hardware functions
 ***********************************************************************/

static void
mdio_write(struct eth_device *dev, int phy_id, int location, unsigned short val)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;

	/* wait for busy */
	while (candy_in(p->mind) & BUSY)
		;

	/* start the write */

	candy_out(p->madr,
		  ((phy_id << FIAD_SHIFT) & FIAD_MASK) |
		  (location & RGAD_MASK));
	candy_out(p->mwtd, (unsigned long)val);
}

static unsigned short
mdio_read(struct eth_device *dev, int phy_id, int location)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;

	/* wait for busy */
	while (candy_in(p->mind) & BUSY)
		;

	candy_out(p->madr,
		  ((phy_id << FIAD_SHIFT) & FIAD_MASK) |
		  (location & RGAD_MASK));
	candy_out(p->mcmd, RSTAT);

	/* wait for busy */
	while (candy_in(p->mind) & BUSY)
		;

	return (unsigned short)candy_in(p->mrdd);
}

static void
candy_set_media_speed(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	int lpa, advertise, media, reg;
	ulong ccr;

	/*
	 * As described in specs, SDP100 works with RMII mode, and is not
	 * related to MII mode. Hence if MII selected, let's leave it ON.
	 * this will be a workaround for VR4133A rev 2.0, formerly known as
	 * WORKAROUND_E19_VR4133A.
	 *
	 * SPD100   10HALF  10FULL  100HALF 100FULL
	 * mii      1       1       1       1
	 * rmii     1       1       0       0
	 */

	/* load default settings */
	ccr = SPD100 | MII_PIN_SELECT;
	if (candy_rmii)
		ccr |= RMII_MODE;

	reg = mdio_read(dev, pp->phy_id, PHY_BMCR);
	if (!(reg & PHY_BMCR_AUTON))
		lpa = advertise = 0;
	else {
		lpa = mdio_read(dev, pp->phy_id, PHY_ANLPAR);
		advertise = mdio_read(dev, pp->phy_id, PHY_ANAR);
	}

	/*
	 * If auto-negotiation is disabled or something weird is there,
	 * conform to the control register.
	 */
	if (lpa == 0 || advertise == 0) {
		pp->link_status.fullduplex = (reg & PHY_BMCR_DPLX) ? 1 : 0;
		if (reg & PHY_BMCR_100MB) {
			if (candy_rmii)
				ccr &= ~SPD100;
			pp->link_status.speed100 = 1;
		} else {
			pp->link_status.speed100 = 0;
		}
		candy_out(pp->regs->ccr, ccr);
		return;
	}

	/* mii_nway_result */
	if ((lpa & advertise) & PHY_ANLPAR_TXFD)
		media = PHY_ANLPAR_TXFD;
	else if ((lpa & advertise) & PHY_ANLPAR_T4)
		media = PHY_ANLPAR_T4;
	else if ((lpa & advertise) & PHY_ANLPAR_TX)
		media = PHY_ANLPAR_TX;
	else if ((lpa & advertise) & PHY_ANLPAR_10FD)
		media = PHY_ANLPAR_10FD;
	else
		media = PHY_ANLPAR_10;

	switch (media) {
	case PHY_ANLPAR_10FD:
	case PHY_ANLPAR_10:
		pp->link_status.speed100 = 0;
		break;
	default:
		if (candy_rmii)
			ccr &= ~SPD100;
		pp->link_status.speed100 = 1;
	}

	switch (media) {
	case PHY_ANLPAR_TX:
	case PHY_ANLPAR_10:
		pp->link_status.fullduplex = 0;
		break;
	default:
		pp->link_status.fullduplex = 1;
	}

	candy_out(pp->regs->ccr, ccr);
}

static int set_mac_addr(struct eth_device *dev, unsigned char *addr)
{
	candy_regs *p = ((struct candy_private *)dev->priv)->regs;

	candy_out(p->lsa2, (addr[0] << 8 | addr[1]) & LSA2_MASK);
	candy_out(p->lsa1,
		  addr[2] << 24 | addr[3] << 16 | addr[4] << 8 | addr[5]);

	return 0;
}

static void candy_filter_init(struct eth_device *dev)
{
	candy_regs *p = ((struct candy_private *)dev->priv)->regs;

	candy_out(p->ht1, 0);
	candy_out(p->ht2, 0);

	candy_out(p->afr, ABC); /* unicast + broadcast */
}

/*
 * This is called when system boots up and/or after ether chip is reset.
 */
static int candy_hw_init(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;
	unsigned int temp;

	candy_out(p->ccr, SRT);

	/*
	 * MII software reset
	 */
	/* we need to maintain CLKS[2:0] = 3'b001 during reset */
	candy_out(p->miic, MISRT | CLKS33);
	candy_out(p->miic, CLKS33);
	/* wait for the reset signal to be propagated through FFs */
	udelay(3);

	/*
	 * initialize MAC
	 */
	candy_out(p->macc1, TXFC | RXFC | CRCEN | PADEN);
	candy_out(p->macc2, APD);
	candy_out(p->ipgt, IPGT);
	candy_out(p->ipgr, IPGR1 | IPGR2);
	candy_out(p->clrt, LCOLW | RETRY);
	candy_out(p->lmax, MAXF);
	candy_out(p->vltp, 0x0);

	switch (MIIC_CLK) {
	case MIIC_CLKS166:
		candy_out(p->miic, CLKS166);
		break;
	case MIIC_CLKS133:
		candy_out(p->miic, CLKS133);
		break;
	case MIIC_CLKS100:
		candy_out(p->miic, CLKS100);
		break;
	case MIIC_CLKS66:
		candy_out(p->miic, CLKS66);
		break;
	case MIIC_CLKS50:
		candy_out(p->miic, CLKS50);
		break;
	default:
		candy_out(p->miic, CLKS33);
	}

	/*
	 * initialize DMA / FIFO
	 */
	candy_out(p->txcfg, DTBS16);
	candy_out(p->txfc, (TPTV | TX_DRTH | TX_FLTH32W));
	candy_out(p->rxcfg, DRBS16);
	candy_out(p->rxfc, (UWM | LWM | RX_DRTH16W));
	candy_out(p->rxpd, AL | DUWM | DLWM);

	/*
	 * mode configuration
	 */
	temp = 0x00800000; /* RSTH */
	candy_out(p->mode, temp);

	/* Set pins for MII/RMII use */
	candy_set_media_speed(dev);

	if (pp->link_status.fullduplex)
		candy_set_macc1_bits(p->macc1, FULLD);
	else
		candy_clear_macc1_bits(p->macc1, FULLD);

	/*
	 * MAC software reset - set/clear these bits all together
	 */
	candy_set_macc2_bits(p->macc2, (MCRST | RFRST | TFRST));
	udelay(3);
	candy_clear_macc2_bits(p->macc2, (MCRST | RFRST | TFRST));

	return 0;
}

static void
candy_down(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;

	candy_clear_macc1_bits(p->macc1, SRXEN);

	candy_clear_bits(p->txcfg, TXE);
	candy_clear_bits(p->rxcfg, RXE);
}

static void
candy_up(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;

	/* set number of descriptors */
	candy_out(p->rxpd, AL | (RX_RING_SIZE & RNOD_MASK));

	/* set the receive descriptor pointer */
	candy_out(p->rxdp, CPHYSADDR(&pp->rx_ring[pp->rx_head]));

	/* enable transmit and receive */
	candy_set_bits(p->txcfg, TXE);
	candy_set_bits(p->rxcfg, RXE);

	/* start receiving */
	candy_set_macc1_bits(p->macc1, SRXEN);
}

/***********************************************************************
 * hardware-independent helper routine
 ***********************************************************************/
static void candy_init_rings(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	int i;

	debug("tx_ring:0x%08p, rx_ring:0x%08p\n", pp->tx_ring, pp->rx_ring);

	for (i = 0; i < RX_RING_SIZE; i++) {
		pp->rx_ring[i].status = DB_LP;
		pp->rx_ring[i].size = PKTSIZE_ALIGN;
		pp->rx_ring[i].pointer = CPHYSADDR(NetRxPackets[i]);
		flush_cache((unsigned long)NetRxPackets[i], PKTSIZE_ALIGN);
	}
	pp->rx_ring[RX_RING_SIZE].status = 0;	/* link back to the beginning */
	pp->rx_ring[RX_RING_SIZE].size = 0xffff;
	pp->rx_ring[RX_RING_SIZE].pointer = CPHYSADDR(pp->rx_ring);

	for (i = 0; i < TX_RING_SIZE; i++) {
		pp->tx_ring[i].size = 0;
		pp->tx_ring[i].status = 0;
		pp->tx_ring[i].pointer = 0;
	}
	pp->tx_ring[TX_RING_SIZE].status = 0;
	pp->tx_ring[TX_RING_SIZE].size = 0x0;
	pp->tx_ring[TX_RING_SIZE].pointer = CPHYSADDR(pp->tx_ring);

	pp->rx_head = 0;
}

static void reclaim_one_rx_desc(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *) dev->priv;
	candy_regs *p = pp->regs;
	volatile struct candy_desc *dp = &pp->rx_ring[pp->rx_head];

	/*
	 * dp->pointer always holds CPHYSADDR(NetRxPackets[pp->rx_head]).
	 * we don't have to update or initialize dp->pointer.
	 */

	dp->status = DB_LP;	/* 1 stands for buffer vs link ptr */
	dp->size = PKTSIZE_ALIGN;

	/* we need to clean up cache here.  Otherwise we may have some
	 * dirty cache while ether controller is feeding fresh pkt data
	 * to the physical RAM.  Data corruption could happen.
	 */
	flush_cache(KSEG0ADDR(dp->pointer), PKTSIZE_ALIGN);

	if (++pp->rx_head == RX_RING_SIZE)
		pp->rx_head = 0;

	/* tell hardware we have one descriptor to work with */
	candy_out(p->rxpd, AL | 1);
}

/***********************************************************************
 * high-level u-boot-related functions
 ***********************************************************************/

/**************************************************************************/
/* INIT - Look for an adapter, this routine's visible to the outside      */
/**************************************************************************/
static int candy_init(struct eth_device *dev, bd_t *bd)
{
	candy_init_rings(dev);

	candy_hw_init(dev);

	set_mac_addr(dev, dev->enetaddr);

	candy_filter_init(dev);

	candy_up(dev);

	return 1;
}

/**************************************************************************/
/* HALT - Turn off ethernet interface                                     */
/**************************************************************************/
static void candy_halt(struct eth_device *dev)
{
	candy_down(dev);
}

/**************************************************************************/
/* SEND - Transmit a frame                                                */
/***************************************************************************/
static int candy_send(struct eth_device *dev, volatile void *packet, int length)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	candy_regs *p = pp->regs;
	int i;
	int res;

	/* tx data buffer must be writeback */
	flush_cache((unsigned long)packet, length);

	/* tx descriptor should always be idle */
	pp->tx_ring[0].size = length;
	pp->tx_ring[0].status = LAST | DB_LP;
	pp->tx_ring[0].pointer = (virt_to_phys(packet));
	pp->tx_ring[1].size = 0;	/* NULL descriptor */
	pp->tx_ring[1].status = 0;
	pp->tx_ring[1].pointer = 0;

	/*
	 * FIXME: This problem is associated with system bus.
	 *
	 * Now we've added a NULL descriptor and are about to re-start xmit.
	 * Beware of tx descs posted writes.
	 */
	MEMSYNC(pp->tx_ring);	/* read wherever, so long as uncached */

	/* Starting Xmit... */
	candy_out(p->txdp, (virt_to_phys(pp->tx_ring)));

	/* wait Tx Done */
	i = 0;
	while (!(pp->tx_ring[0].status & OWN)) {
		if (i > MAX_WAIT) {
			printf("TX timeout:%x\n", (pp->tx_ring[0].status));
			break;
		}
		udelay(10);
		i++;
	}

	res = pp->tx_ring[0].status;
	return res;
}

/**************************************************************************/
/* RECV - Receive a frame                                                 */
/**************************************************************************/
/***** data output from now rx-descriptor to data buffer *****/
/* return value
 *  0: receive is OK
 *  1: nothing has been received
 *  2: receive data is ERROR
 */
static int candy_recv(struct eth_device *dev)
{
	struct candy_private *pp = (struct candy_private *)dev->priv;
	volatile struct candy_desc *dp;
	int i;

	for (i = 0; ; i++) {
		dp = &pp->rx_ring[pp->rx_head];

		/* nothing has been received */
		if (!(dp->status & OWN))
			break;

		/* receive OK */
		if (!(dp->status & RXOK)) {
			printf("[Rx Error] status:%04x\n", dp->status);
			reclaim_one_rx_desc(dev);
			continue;
		}

		/* Pass the packet up to the protocol layyers */
		NetReceive(NetRxPackets[pp->rx_head], dp->size - 4);

		reclaim_one_rx_desc(dev);
	}
	return 0;
}

#ifdef CONFIG_CMD_MII
/*
 * Check if given phyaddr is valid, i.e. there is a PHY connected.
 * Do this by checking model value field from ID2 register.
 */
static struct eth_device *verify_phyaddr(const char *devname, unsigned char addr)
{
	struct eth_device *dev;
	unsigned short value;
	unsigned char model;

	dev = eth_get_dev_by_name(devname);
	if (dev == NULL) {
		printf("%s: no such device\n", devname);
		return NULL;
	}

	/* read id2 register */
	value = mdio_read(dev, addr, PHY_PHYIDR2);

	/* get model */
	model = (unsigned char)((value >> 4) & 0x003f);

	if ((model == 0x3f) || (model == 0)) {
		printf("%s: no PHY at address %d\n", devname, addr);
		return NULL;
	}

	return dev;
}

static int candy_miiphy_write(const char *devname, unsigned char addr,
			      unsigned char reg, unsigned short value)
{
	struct eth_device *dev;

	dev = verify_phyaddr(devname, addr);
	if (dev == NULL)
		return -1;

	mdio_write(dev, addr, reg, value);
	return 0;
}

static int candy_miiphy_read(const char *devname, unsigned char addr,
			     unsigned char reg, unsigned short *value)
{
	struct eth_device *dev;

	dev = verify_phyaddr(devname, addr);
	if (dev == NULL)
		return -1;

	*value = mdio_read(dev, addr, reg);
	return 0;
}
#endif /* CONFIG_CMD_MII */

static int register_candy(void *hw_addr, int phy_id, unsigned char *mac,
			  const char *post, bd_t *bis)
{
	struct eth_device *dev;
	struct candy_private *pp;
	int i;

	dev = (struct eth_device *)malloc(sizeof(*dev));
	memset(dev, 0, sizeof(*dev));
	pp = (struct candy_private *)malloc(sizeof(*pp));
	memset(pp, 0, sizeof(*pp));

	sprintf(dev->name, "NEC-Candy%s", post);
	dev->iobase = 0;
	dev->priv = pp;
	dev->init = candy_init;
	dev->halt = candy_halt;
	dev->send = candy_send;
	dev->recv = candy_recv;
	for (i = 0; i < 6; ++i)
		dev->enetaddr[i] = mac[i];

	pp->regs = hw_addr;

	/* alloc tx/rx rings and rx buffers */
	flush_cache((unsigned long)tx_descs,
		    sizeof(struct candy_desc) * (TX_RING_SIZE + 1));
	flush_cache((unsigned long)rx_descs,
		    sizeof(struct candy_desc) * (RX_RING_SIZE + 1));
	pp->tx_ring = (void *) KSEG1ADDR(&tx_descs[0]);
	pp->rx_ring = (void *) KSEG1ADDR(&rx_descs[0]);

	pp->phy_id = phy_id;
	eth_register(dev);

#ifdef CONFIG_CMD_MII
	miiphy_register(dev->name, candy_miiphy_read, candy_miiphy_write);
#endif

	debug("Probe candy chip at : 0x%08lx, Phy ID 0x%04x, "
	      "MAC Addr %02x:%02x:%02x:%02x:%02x:%02x\n",
	      pp->regs, pp->phy_id,
	      dev->enetaddr[0], dev->enetaddr[1], dev->enetaddr[2],
	      dev->enetaddr[3], dev->enetaddr[4], dev->enetaddr[5]);

	return 1;
}

int candy_enet_initialize(bd_t *bis)
{
	unsigned char bi_enetaddr[6];
	eth_getenv_enetaddr("ethaddr", bi_enetaddr);
	return register_candy((void *)ETH0_BASE, ETH0_PHY_ADDR,
			      ((unsigned char *)bi_enetaddr), "", bis);
}

#endif /* CONFIG_NEC_CANDY */
