/*
 * emma_spi.c
 *
 * SPI driver for the EMMA3Sx BROI Serial Flash control.
 *
 * Copyright (C) 2011 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include <common.h>
#include <asm/io.h>
#include <asm/arch-emma/emma.h>
#include <malloc.h>
#include <spi.h>

#define EMMA_SF_ACCESS_MODE	(0x5c)
#define EMMA_SF_CONTROL		(0x60)
#define EMMA_SF_ADDRESS		(0x64)
#define EMMA_SF_DATA_LENGTH	(0x68)
#define EMMA_SF_DATA_COUNT	(0x6c)

#if defined(CONFIG_EMMA3SEP)
#define OTHER_ROM_IF_PORT	(0x00004000)
#define SF_ROM_IF_PORT		(0x00010000)
#elif defined(CONFIG_EMMA3SL2LP)
#define OTHER_ROM_IF_PORT	(0x00004000)
#define SF_ROM_IF_PORT		(0x00010000)
#elif defined(CONFIG_EMMA3SLx)
#define OTHER_ROM_IF_PORT	(0x00008000)
#define SF_ROM_IF_PORT		(0x00020000)
#endif

struct emma_spi_slave {
	struct spi_slave slave;
	ulong base;
};
#define to_emma_spi_slave(s) container_of(s, struct emma_spi_slave, slave)

static u32 emma_sf_address;
static u32 save_alternate;

static u32 sf_base_address_table[6][4] = {
	 /*   CS0         CS1         CS2         CS3  */	/* mode */
	{ 0xBFC00000, 0xBF800000, 0xBF400000, 0xBF000000 },	/* 000	*/
	{ 0xBF800000, 0xBF000000, 0xBE800000, 0xBE000000 },	/* 001	*/
	{ 0xBF000000, 0xBE000000, 0xBD000000, 0xBC000000 },	/* 010	*/
	{ 0xBE000000, 0xBC000000, 0xFFFFFFFF, 0xFFFFFFFF },	/* 011	*/
	{ 0xBC000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF },	/* 100	*/
	{ 0xBC000000, 0xBC000000, 0xBC000000, 0xBC000000 }	/* 101	*/
};

__attribute__((weak))
int spi_cs_is_valid(unsigned int bus, unsigned int cs)
{
	u32 reg;

	reg = emma_readl(BROI_ROM_MODE) >> 16;

	if (((reg >> (cs * 2)) & 0x03) == 0x03)
		return 1;
	else
		return 0;
}

__attribute__((weak))
void spi_cs_activate(struct spi_slave *slave)
{
}

__attribute__((weak))
void spi_cs_deactivate(struct spi_slave *slave)
{
}

void spi_init()
{
	/* unused */
}

struct spi_slave *spi_setup_slave(unsigned int bus, unsigned int cs,
			unsigned int max_hz, unsigned int mode)
{
	struct emma_spi_slave	*espi;

	if (!spi_cs_is_valid(bus, cs))
		return NULL;

	espi = malloc(sizeof(struct emma_spi_slave));
	if (!espi)
		return NULL;

	espi->slave.bus = bus;
	espi->slave.cs = cs;
	espi->base = KSEG1ADDR(BROI_BASE);
	debug("%s: bus:%i cs:%i base:%lx\n", __func__,
		bus, cs, espi->base);

	emma_sf_address = sf_base_address_table[0][0];

	return &espi->slave;
}

void spi_free_slave(struct spi_slave *slave)
{
	struct emma_spi_slave *espi = to_emma_spi_slave(slave);

	free(espi);
}

int spi_claim_bus(struct spi_slave *slave)
{
	struct emma_spi_slave *espi = to_emma_spi_slave(slave);
	u32 reg;

	debug("%s: bus:%i cs:%i\n", __func__, slave->bus, slave->cs);

	reg = emma_readl(BHIF_ALTERNATE0);
	save_alternate = reg & OTHER_ROM_IF_PORT;
	reg &= ~save_alternate;		/* clear other ROM I/F */
	reg |= SF_ROM_IF_PORT;		/* select SF-ROM I/F */
	emma_writel(reg, BHIF_ALTERNATE0);

	reg = emma_readl(BROI_ROM_MODE) & 0x07;
	emma_sf_address = sf_base_address_table[reg][slave->cs];

	writel(0x00010000, espi->base + EMMA_SF_ACCESS_MODE);
	udelay(10);

	return 0;
}

void spi_release_bus(struct spi_slave *slave)
{
	struct emma_spi_slave *espi = to_emma_spi_slave(slave);
	u32 reg;

	debug("%s: bus:%i cs:%i\n", __func__, slave->bus, slave->cs);

	writel(0x00010001, espi->base + EMMA_SF_ACCESS_MODE);	/* to single access mode */
	udelay(10);

	reg = emma_readl(BHIF_ALTERNATE0);
	reg &= ~SF_ROM_IF_PORT;		/* clear SF-ROM I/F */
	reg |= save_alternate;		/* select ROM I/F */
	emma_writel(reg, BHIF_ALTERNATE0);
}

int spi_xfer(struct spi_slave *slave, unsigned int bitlen,
		const void *dout, void *din, unsigned long flags)
{
	struct emma_spi_slave *espi = to_emma_spi_slave(slave);
	/* assume spi core configured to do 8 bit transfers */
	uint bytes = bitlen / 8;
	const uchar *txp = dout;
	uchar *rxp = din;
	uint addr, control;
	uint data;

	debug("%s: bus:%i cs:%i bitlen:%i bytes:%i flags:%lx\n", __func__,
		slave->bus, slave->cs, bitlen, bytes, flags);
	if (bitlen == 0)
		goto done;

	if (bitlen % 8) {
		flags |= SPI_XFER_END;
		goto done;
	}

	if (flags & SPI_XFER_BEGIN)
		spi_cs_activate(slave);

	if (txp && !rxp && (bytes <= 5)) {
		/* command cycle */
		control = (txp[0] << 24);
		if (bytes >= 4) {
			addr = (txp[1] << 16) | (txp[2] << 8) | txp[3];
			debug("emma_sf: write address reg:%08x\n", addr);
			writel(addr, espi->base + EMMA_SF_ADDRESS);
			/* set up the address cycle : 24bits */
			control |= (24 << 16);
			if (bytes >= 5) {
				/* set up the dummy cycle : 8bits */
				control |= (8 << 8);
			}
		}
		control |= (espi->slave.cs << 2);
		if (flags & SPI_XFER_END)
			control |= 0x01;	/* immediate */
		else
			control |= 0x03;	/* non-immediate */
		debug("emma_sf: write control reg:%08x\n", control);
		writel(control, espi->base + EMMA_SF_CONTROL);
	} else {
		/* data transfer cycle */
		debug("emma_sf: write datalen reg:%08x\n", bytes);
		if (txp)
			writel(bytes, espi->base + EMMA_SF_DATA_LENGTH);
		else
			writel(65536, espi->base + EMMA_SF_DATA_LENGTH);

		while (bytes) {
			if (txp) {
				data = (*txp++) << 24;
				data |= (*txp++) << 16;
				data |= (*txp++) << 8;
				data |= (*txp++);
				writel(data, emma_sf_address);
			}
			if (rxp) {
				data = readl(emma_sf_address);
				*rxp++ = data >> 24;
				*rxp++ = data >> 16;
				*rxp++ = data >> 8;
				*rxp++ = data;
			}
			bytes = (bytes > 4) ? (bytes - 4) : 0;
		}
	}

 done:
	if (flags & SPI_XFER_END) {
		if ((readl(espi->base + EMMA_SF_DATA_LENGTH) & 0x0001ffff) == 0x00010000) {
			udelay(10);
			writel(0, espi->base + EMMA_SF_CONTROL);
		}
		while((readl(espi->base + EMMA_SF_DATA_COUNT) & 0xf0000000) != 0xf0000000);
		spi_cs_deactivate(slave);
	}

	return 0;
}
